/** @file TherionLine.h
 *
 * @author marco corvi
 * @date aug. 2009
 *
 * @brief plot therion line
 * --------------------------------------------------------
 *  Copyright This sowftare is distributed under GPL-3.0 or later
 *  See the file COPYING.
 */
#ifndef THERION_LINE_H
#define THERION_LINE_H

#include <math.h> // fabs
#include <vector>

#include "ThLineType.h"
#include "TherionPoint.h"

/** A line point.
 * A line point vcan have two control points, Point-1 before it and Point-2
 * after it. When saving the therion file the control points should be used
 * as follow:
 *   - the first point has no control points, so write
 *     x0 y0
 *   - if the previous point or the current point have control points write
 *     x2_prev y2_prev x1 y1 x0, y0
 *     (in case one of the two points does not have control points, use the
 *     point coords for the control point coords)
 *   - if neither the previous point nor the current point have control points write
 *     x0 y0
 * FIXME i have to find out how to write for closed lines at start/end
 */
struct ThLinePoint : public ThPoint
{
  double x1, y1;
  double x2, y2;
  bool has1, has2;

  ThLinePoint( double x, double y )
    : ThPoint( x, y )
    , has1( false )
    , has2( false )
  { }

  void setControl1( double x10, double y10 )
  {
    if ( fabs(x10-x) > 2.0 || fabs(y10-y) > 2.0 ) {
      x1 = x10;
      y1 = y10;
      has1 = true;
    }
  }

  void setControl2( double x20, double y20 )
  {
    if ( fabs(x20-x) > 2.0 || fabs(y20-y) > 2.0 ) {
      x2 = x20;
      y2 = y20;
      has2 = true;
    }
  }

  void setControls( double x20, double y20 )
  {
    if ( fabs(x20-x) > 2.0 || fabs(y20-y) > 2.0 ) {
      x1 = 2*x - x20;
      y1 = 2*y - y20;
      x2 = x20;
      y2 = y20;
      has1 = has2 = true;
      printf("LP %.2f %.2f %.2f %.2f %.2f %.2f\n", x, y, x1, y1, x2, y2 );
    }
  }

  void clearControl1() { has1 = false; }

  void clearControl2() { has2 = false; }
};

/** plot therion line
 */
struct ThLine 
{
  typedef std::vector< ThLinePoint >::iterator iterator;
  typedef std::vector< ThLinePoint >::const_iterator const_iterator;
 
  ThLineType type;            //!< line type
  bool closed;                //!< whether the line is closed
  std::vector< ThLinePoint > pts; //!< line control points

  /** cstr
   * @param t  line type (default THL_USER)
   */
  ThLine( ThLineType t = THL_USER )
    : type( t )
    , closed( false )
  { }

  /** accessor: get the number of control points
   * @return the size of the vector of points
   */
  size_t Size() const { return pts.size(); }

  /** set the line close
   * @param close   value of closed
   */
  void setClosed( bool close ) { closed = close; }

  /** check if the line is closed
   * @return true if the lin eis closed
   */
  bool isClosed() const { return closed; }

  /** append a point
   * @param x   point X coord.
   * @param y   point Y coord.
   * @return pointer to the last inserted point
   */
  ThLinePoint * Add( double x, double y ) 
  { 
    pts.push_back( ThLinePoint(x,y) ); 
    return &( *pts.rbegin() );
  }

  /** erase the last point;
   */
  void DropLast() 
  { 
    iterator it = pts.end();
    --it;
    pts.erase( it );
  }

  iterator Begin() { return pts.begin(); }
  const_iterator Begin() const { return pts.begin(); }

  iterator End() { return pts.end(); }
  const_iterator End() const { return pts.end(); }

  ThPoint & operator[]( int k ) { return pts[k]; }

};

#endif

